from ..event.base import BaseEvent
from ..app.base import BaseApp


__all__ = [
    'BaseCommand',
]


class BaseCommand(BaseEvent):
    async def handle(self, app: BaseApp, corr_id: str):
        await self.execute(app=app, corr_id=corr_id)

    async def execute(self, app: BaseApp, corr_id: str):
        pass


class BaseCommandResult(BaseEvent):
    @property
    def error(self):
        return self.data.get('error')

    @property
    def error_label(self):
        if self.data['error']:
            return self.data['error']['label']
        else:
            return None

    @property
    def error_message(self):
        if self.data['error']:
            return self.data['error'].get('message')
        else:
            return None



# coding: utf-8
# Copyright (c) Qotto, 2018-2020

import asyncio
import logging
import sys

from aiokafka import AIOKafkaConsumer
from typing import Any, Dict, List

from ..app.base import BaseApp
from ..serializer.base import BaseEventSerializer
from .base import BaseEventConsumer

__all__ = [
    'KafkaConsumer'
]

logger = logging.getLogger(__name__)


class KafkaConsumer(BaseEventConsumer):
    def __init__(
        self,
        settings: object,
        app: BaseApp,
        serializer: BaseEventSerializer,
        event_topics: List[str],
        event_group: str,
        position: str,
    ) -> None:
        if not hasattr(settings, 'KAFKA_BOOTSTRAP_SERVER'):
            raise Exception('Missing KAFKA_BOOTSTRAP_SERVER config')

        self.max_retries = 10
        if hasattr(settings, 'EVENTY_CONSUMER_MAX_RETRIES'):
            self.max_retries = settings.EVENTY_CONSUMER_MAX_RETRIES

        self.retry_interval = 1000
        if hasattr(settings, 'EVENTY_CONSUMER_RETRY_INTERVAL'):
            self.retry_interval = settings.EVENTY_CONSUMER_RETRY_INTERVAL

        self.retry_backoff_coeff = 2
        if hasattr(settings, 'EVENTY_CONSUMER_RETRY_BACKOFF_COEFF'):
            self.retry_backoff_coeff = settings.EVENTY_CONSUMER_RETRY_BACKOFF_COEFF

        self.app = app
        self.event_topics = event_topics
        self.event_group = event_group
        self.position = position
        self.consumer = None
        self.current_position_checkpoint_callback = None
        self.end_position_checkpoint_callback = None
        bootstrap_servers = settings.KAFKA_BOOTSTRAP_SERVER

        consumer_args: Dict[str, Any]
        consumer_args = {
            'loop': asyncio.get_event_loop(),
            'bootstrap_servers': [bootstrap_servers],
            'enable_auto_commit': False,
            'group_id': self.event_group,
            'value_deserializer': serializer.decode,
            'auto_offset_reset': self.position
        }

        try:
            self.consumer = AIOKafkaConsumer(
                *self.event_topics, **consumer_args)

        except Exception as e:
            logger.error(
                f"Unable to connect to the Kafka broker {bootstrap_servers} : {e}")
            raise e

    def set_current_position_checkpoint_callback(self, checkpoint_callback):
        self.current_position_checkpoint_callback = checkpoint_callback

    def set_end_position_checkpoint_callback(self, checkpoint_callback):
        self.end_position_checkpoint_callback = checkpoint_callback

    async def current_position(self):
        # Warning: this method returns last committed offsets for each assigned partition
        position = {}
        for partition in self.consumer.assignment():
            offset = await self.consumer.committed(partition) or 0
            position[partition] = offset
        return position

    async def consumer_position(self):
        # Warning: this method returns current offsets for each assigned partition
        position = {}
        for partition in self.consumer.assignment():
            position[partition] = await self.consumer.position(partition)
        return position

    async def end_position(self):
        position = {}
        for partition in self.consumer.assignment():
            offset = (await self.consumer.end_offsets([partition]))[partition]
            position[partition] = offset
        return position

    async def is_checkpoint_reached(self, checkpoint):
        for partition in self.consumer.assignment():
            position = (await self.consumer.position(partition))
            if position < checkpoint[partition]:
                return False
        return True

    async def start(self):
        logger.info(
            f'Starting kafka consumer on topic {self.event_topics} with group {self.event_group}'
        )
        try:
            await self.consumer.start()
        except Exception as e:
            logger.error(
                f'An error occurred while starting kafka consumer '
                f'on topic {self.event_topics} with group {self.event_group}: {e}'
            )
            sys.exit(1)

        current_position_checkpoint = None
        end_position_checkpoint = None
        if self.event_group is not None:
            current_position = await self.current_position()
            end_position = await self.end_position()
            logger.debug(f'Current position : {current_position}')
            logger.debug(f'End position : {end_position}')

            if self.position == 'earliest' and self.event_group is not None:
                current_position_checkpoint = current_position
                end_position_checkpoint = end_position
                await self.consumer.seek_to_beginning()

        async for msg in self.consumer:
            retries = 0
            sleep_duration_in_ms = self.retry_interval
            while True:
                try:
                    event = msg.value
                    corr_id = event.correlation_id

                    logger.info(
                        f"[CID:{corr_id}] Start handling {event.name}")
                    await event.handle(app=self.app, corr_id=corr_id)
                    logger.info(
                        f"[CID:{corr_id}] End handling {event.name}")

                    if self.event_group is not None:
                        logger.debug(
                            f"[CID:{corr_id}] Commit Kafka transaction")
                        await self.consumer.commit()

                    logger.debug(
                        f"[CID:{corr_id}] Continue with the next message")
                    # break the retry loop
                    break
                except Exception:
                    logger.exception(
                        f'[CID:{corr_id}] An error occurred while handling received message.'
                    )

                    if retries != self.max_retries:
                        # increase the number of retries
                        retries = retries + 1

                        sleep_duration_in_s = int(sleep_duration_in_ms / 1000)
                        logger.info(
                            f"[CID:{corr_id}] Sleeping {sleep_duration_in_s}s a before retrying...")
                        await asyncio.sleep(sleep_duration_in_s)

                        # increase the sleep duration
                        sleep_duration_in_ms = sleep_duration_in_ms * self.retry_backoff_coeff

                    else:
                        logger.error(
                            f'[CID:{corr_id}] Unable to handle message within {1 + self.max_retries} tries. Stopping process')
                        sys.exit(1)

            if current_position_checkpoint and await self.is_checkpoint_reached(current_position_checkpoint):
                logger.info('Current position checkpoint reached')
                if self.current_position_checkpoint_callback:
                    await self.current_position_checkpoint_callback()
                current_position_checkpoint = None

            if end_position_checkpoint and await self.is_checkpoint_reached(end_position_checkpoint):
                logger.info('End position checkpoint reached')
                if self.end_position_checkpoint_callback:
                    await self.end_position_checkpoint_callback()
                end_position_checkpoint = None



# coding: utf-8
# Copyright (c) Qotto, 2018

import doctest
import unittest


def load_tests(loader, tests, pattern):
    tests.addTests(doctest.DocTestSuite('eventy.event.base'))
    tests.addTests(doctest.DocTestSuite('eventy.event.generic'))
    tests.addTests(doctest.DocTestSuite('eventy.serializer.base'))
    tests.addTests(doctest.DocTestSuite('eventy.serializer.dummy'))
    tests.addTests(doctest.DocTestSuite('eventy.serializer.avro'))
    tests.addTests(doctest.DocTestSuite('eventy.utils'))
    return tests



import logging

import asyncio
from sanic import Sanic, Blueprint
from sanic.response import json
from typing import Type, Callable, Dict, List

from ..utils import load_class
from ..command.base import BaseCommand
from ..event.base import BaseEvent
from ..serializer.base import BaseEventSerializer
from ..emitter.base import BaseEventEmitter
from ..consumer.base import BaseEventConsumer
from .base import BaseApp

health_bp = Blueprint('health')


@health_bp.route('/health')
async def health(request):
    return json({}, status=200)


class Eventy(BaseApp):
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.context: Dict[str, object] = dict()
        self.http_handler = None
        self.http_handler_port = None

    def start(self):
        if self.http_handler is not None:
            asyncio.ensure_future(self.http_handler.create_server(
                host="0.0.0.0", port=self.http_handler_port, debug=self.http_debug, access_log=self.http_access_log))

        asyncio.get_event_loop().run_forever()

    def configure_event_serializer(self, settings: object, serializer_name: str) -> BaseEventSerializer:
        if not hasattr(settings, 'EVENTY_EVENT_SERIALIZER'):
            raise Exception('Missing EVENTY_EVENT_SERIALIZER config')
        self.logger.debug(
            f"Configure serializer {settings.EVENTY_EVENT_SERIALIZER}")
        serializer_class = load_class(settings.EVENTY_EVENT_SERIALIZER)
        serializer = serializer_class(settings=settings)
        self.set(serializer_name, serializer)
        return serializer

    def configure_consumer(self, settings: object, serializer: BaseEventSerializer, consumer_name: str, event_topics: List[str], event_group: str, position: str)-> BaseEventConsumer:
        if not hasattr(settings, 'EVENTY_EVENT_CONSUMER'):
            raise Exception('Missing EVENTY_EVENT_CONSUMER config')
        self.logger.debug(
            f"Initializing consumer {settings.EVENTY_EVENT_CONSUMER}")
        consumer_class = load_class(
            settings.EVENTY_EVENT_CONSUMER)
        consumer = consumer_class(settings=settings, app=self,
                                  serializer=serializer,  event_topics=event_topics, event_group=event_group, position=position)
        asyncio.ensure_future(consumer.start())
        self.set(consumer_name, consumer)
        return consumer

    def configure_emitter(self, settings: object, serializer: BaseEventSerializer, emitter_name: str)->BaseEventEmitter:
        if not hasattr(settings, 'EVENTY_EVENT_EMITTER'):
            raise Exception('Missing EVENTY_EVENT_EMITTER config')
        self.logger.debug(
            f"Initializing emitter {settings.EVENTY_EVENT_EMITTER}")
        emitter_class = load_class(
            settings.EVENTY_EVENT_EMITTER)
        emitter = emitter_class(settings=settings,
                                serializer=serializer)
        self.set(emitter_name, emitter)
        return emitter

    def configure_http_handler(self, settings: object, http_handler_name: str) -> Sanic:
        if hasattr(settings, 'EVENTY_HTTP_HANDLER_PORT'):
            self.http_handler_port = settings.EVENTY_HTTP_HANDLER_PORT
        else:
            self.http_handler_port = 8000

        if hasattr(settings, 'EVENTY_HTTP_ACCESS_LOG'):
            self.http_access_log = settings.EVENTY_HTTP_ACCESS_LOG
        else:
            self.http_access_log = False

        if hasattr(settings, 'EVENTY_HTTP_DEBUG'):
            self.http_debug = settings.EVENTY_HTTP_DEBUG
        else:
            self.http_debug = False

        self.logger.debug(
            f"Initializing http handler on port {self.http_handler_port}")
        self.http_handler = Sanic()

        self.http_handler.request_middleware.append(
            self.attach_app_to_http_request)

        self.http_handler.blueprint(health_bp)

        self.set(http_handler_name, self.http_handler)

        return self.http_handler

    def attach_app_to_http_request(self, request):
        request['app'] = self

    def set(self, object_name: str, object_instance: object):
        self.context[object_name] = object_instance

    def get(self, object_name: str):
        return self.context.get(object_name)



# coding: utf-8
# Copyright (c) Qotto, 2018

from ..event.base import BaseEvent
from .base import BaseEventSerializer

__all__ = [
    'DummyEventSerializer',
]


class DummyEventSerializer(BaseEventSerializer):
    def encode(self, event: BaseEvent) -> BaseEvent:
        return event

    def decode(self, event: BaseEvent) -> BaseEvent:
        return event



# coding: utf-8
# Copyright (c) Qotto, 2018

from .base import BaseEventSerializer
from .dummy import DummyEventSerializer
from .avro import AvroEventSerializer

__all__ = [
    'BaseEventSerializer',
    'DummyEventSerializer',
    'AvroEventSerializer',
]



# coding: utf-8
# Copyright (c) Qotto, 2018-2020

from typing import Any, Dict

from ..app.base import BaseApp
from ..utils import current_timestamp

__all__ = [
    'BaseEvent'
]


class BaseEvent:
    def __init__(self, name: str, data: Dict[str, Any]) -> None:
        if data is None:
            data = dict()

        if 'schema_version' not in data:
            data['schema_version'] = '0.0.0'

        timestamp = current_timestamp()
        if 'event_timestamp' in data:
            timestamp = data['event_timestamp']

        if 'timestamp' in data:
            timestamp = data['timestamp']

        data['timestamp'] = timestamp

        self.data = data
        self.name = name

    async def handle(self, app: BaseApp, corr_id: str):
        pass

    @property
    def context(self):
        return self.data['context']

    @property
    def schema_version(self):
        return self.data['schema_version']

    @property
    def correlation_id(self):
        return self.data['correlation_id']

    @property
    def timestamp(self):
        return self.data['timestamp']

    @classmethod
    def from_data(cls, event_name: str, event_data: Dict[str, Any]):
        raise NotImplementedError



#!/usr/bin/env python

from setuptools import setup, find_packages


if __name__ == '__main__':
    setup(
        name='eventy',
        version='2.5.1',
        url='https://github.com/qotto/eventy',
        license='MIT',
        author='Alexandre Syenchuk',
        author_email='sacha@qotto.net',
        description='Qotto/eventy',
        packages=find_packages(),
        include_package_data=True,
        zip_safe=False,
        install_requires=[
            'avro-python3',
            'pyyaml>=4.2b1',
            'aiokafka==0.5.2',
            'sanic'
        ],
        classifiers=[
            'Intended Audience :: Developers',
            'Programming Language :: Python :: 3.8',
        ],
    )



from aiokafka import AIOKafkaProducer
import logging
from ..event.base import BaseEvent
from ..command.base import BaseCommand
from ..serializer.base import BaseEventSerializer
from typing import Any, Dict
import asyncio
from .base import BaseEventEmitter

__all__ = [
    'KafkaProducer'
]


class KafkaProducer(BaseEventEmitter):

    def __init__(self, settings: object, serializer: BaseEventSerializer)->None:
        self.logger = logging.getLogger(__name__)

        if not hasattr(settings, 'KAFKA_BOOTSTRAP_SERVER'):
            raise Exception('Missing KAFKA_BOOTSTRAP_SERVER config')

        self.serializer = serializer
        self.started = False
        try:
            producer_args: Dict[str, Any]
            producer_args = {
                'loop': asyncio.get_event_loop(),
                'bootstrap_servers': [settings.KAFKA_BOOTSTRAP_SERVER],
                'value_serializer': self.serializer.encode
            }

            self.producer = AIOKafkaProducer(**producer_args)
        except Exception as e:
            self.logger.error(
                f"Unable to connect to the Kafka broker {settings.KAFKA_BOOTSTRAP_SERVER}")
            raise e

    async def send(self, event: BaseEvent, destination: str, key=None):
        if not self.started:
            await self.producer.start()
            self.started = True
        self.logger.debug(
            f'Sending event {event.name} on topic {destination}')
        await self.producer.send_and_wait(destination, event, key)



from ..event.base import BaseEvent

__all__ = [
    'BaseEventEmitter'
]


class BaseEventEmitter:
    def __init__(self, *args, **kw) -> None:
        pass

    async def send(self, event: BaseEvent, destination: str):
        raise NotImplementedError
